from .retries import make_retry_client, retry_default_rule
from .throttle import make_throttled_client
from .dev_cache import (
    make_dev_caching_client,
    MemoryCache,
    FileCache,
    SqliteCache,
    CacheStorageBase,
    _cache_200s,
    _default_keyfunc,
)
from ._types import ResponsePredicate, CacheKeyfunc
from httpx import Client


def make_careful_client(
    *,
    client: Client | None = None,
    retry_attempts: int = 0,
    retry_wait_seconds: float = 10,
    should_retry: ResponsePredicate = retry_default_rule,
    requests_per_minute: int = 0,
    cache_storage: CacheStorageBase = None,
    cache_write_only: bool = False,
    should_cache: ResponsePredicate = _cache_200s,
    cache_keyfunc: CacheKeyfunc = _default_keyfunc,
):
    """
    This function patches an `httpx.Client` so that all requests made with the client support
     [retries](#retries), [throttling](#throttling), and [development caching](#development-caching).


    Parameters:
        client: A pre-configured `httpx.Client`. If omitted a default client will be created.

        retry_attempts: Maximum number of retries. If non-zero will retry up to this many times
                         with increasing wait times, starting with `retry_wait_seconds`.

        retry_wait_seconds: Number of seconds to sleep between first attempt and first retry.
                             Subsequent attempts will increase exponentially (2x, 4x, 8x, etc.)

        should_retry: Predicate function that takes a `httpx.Response` and returns `True` if it should be retried.

        requests_per_minute: Maximum number of requests per minute. (e.g. 30 will throttle to ~2s between requests)

        cache_storage: An object that implements the [cache storage interface](#cache-storage).

        cache_write_only: Update cache, but never read from it.

        should_cache: Predicate function that takes a `httpx.Response` and returns `True` if it should be cached.

        cache_keyfunc: Function that takes request details and returns a unique cache key.

    ## Retries

    If `retry_attempts` is set, responses will be passed to `should_retry`.
    Responses that are rejected (return `True`) will be retried after a wait based on
    `retry_wait_seconds`.
    Each retry will wait twice as long as the one before.

    ## Throttling

    If `requests_per_minute` is set, standard (non-retry) requests will automatically
    sleep for a short period to target the given rate.

    For example, at 30rpm, the sleep time on a fast request will be close to 2 seconds.

    ## Development Caching

    Why **development caching?**

    This feature is named as a reminder that **this is not true HTTP caching**, which
    should take various headers into account. Look at libraries like [hishel](https://hishel.com) if that's what you are after.

    The purpose of this feature is to allow you to cache all of your HTTP requests during development.
    Often when writing a scraper or crawler, you wind up hitting the site you are working on more often than you'd like-- each time you iterate on your code you're likely making redundant requests to pages that haven't changed.

    By caching all successful requests (configurable with the `should_cache` parameter),
    you can easily re-run scrapers without making redundant HTTP requests.
    This means faster development time & happier upstream servers.

    To enable development caching, assign a [`MemoryCache`][careful.httpx.MemoryCache],
    [`FileCache`][careful.httpx.FileCache], or [`SqliteCache`][careful.httpx.SqliteCache] to
    the `cache_storage` property of a `scrapelib.Scraper`.

    ---

    When multiple features are applied, the order of wrapping ensures that:
       - the cache is checked first, and bypasses throttling if hit
       - retries use their own delays, but not throttled separately
    """
    if client is None:
        client = Client()
    # order matters, retry on inside b/c it is last-chance scenario
    if retry_attempts:
        client = make_retry_client(
            client=client,
            attempts=retry_attempts,
            wait_seconds=retry_wait_seconds,
            should_retry=should_retry,
        )
    # throttling around retries
    if requests_per_minute:
        client = make_throttled_client(
            client=client, requests_per_minute=requests_per_minute
        )
    # caching on top layer, so cache will be checked first
    if cache_storage:
        client = make_dev_caching_client(
            client=client,
            cache_storage=cache_storage,
            cache_keyfunc=cache_keyfunc,
            should_cache=should_cache,
            write_only=cache_write_only,
        )

    return client


__all__ = [
    "make_retry_client",
    "make_throttled_client",
    "make_dev_caching_client",
    "MemoryCache",
    "FileCache",
    "SqliteCache",
]



from careful.httpx import make_throttled_client
from unittest import mock
from typing import Any
from httpx import Request
from fakeresponse import FakeResponse


def request_200(request: Request, *args: Any, **kwargs: Any) -> FakeResponse:
    return FakeResponse(request.url, 200, b"ok")


mock_200 = mock.Mock(wraps=request_200)


def test_request_throttling() -> None:
    client = make_throttled_client(requests_per_minute=30)

    mock_sleep = mock.Mock()

    # check that sleep is called on call 2 & 3
    with mock.patch("time.sleep", mock_sleep):
        with mock.patch.object(client, "send", mock_200):
            client.get("http://dummy/")
            client.get("http://dummy/")
            client.get("http://dummy/")
            assert mock_sleep.call_count == 2
            # should have slept for ~2 seconds to aim at 30 per min
            assert 1.8 <= mock_sleep.call_args[0][0] <= 2.2



from pytest_httpbin.serve import Server  # type: ignore
from httpx import Response
from careful.httpx import make_dev_caching_client, MemoryCache, FileCache, SqliteCache


def test_dev_caching(httpbin: Server) -> None:
    client = make_dev_caching_client(cache_storage=MemoryCache(), write_only=False)

    resp = client.get(httpbin.url + "/status/200")
    assert not resp.fromcache
    resp = client.get(httpbin.url + "/status/200")
    assert resp.fromcache


def test_dev_caching_params(httpbin: Server) -> None:
    client = make_dev_caching_client(cache_storage=MemoryCache(), write_only=False)

    resp = client.get(httpbin.url + "/status/200?a=1&b=2")
    assert not resp.fromcache
    resp = client.get(httpbin.url + "/status/200?a=1&b=2")
    assert resp.fromcache
    resp = client.get(httpbin.url + "/status/200?a=1&b=3")
    assert not resp.fromcache


# test storages #####


def _test_cache_storage(storage_obj) -> None:
    # unknown key returns None
    assert storage_obj.get("one") is None

    _content_as_bytes = b"here's unicode: \xe2\x98\x83"
    _content_as_unicode = "here's unicode: \u2603"

    # set 'one'
    resp = Response(200)
    resp.headers["x-num"] = "one"
    resp._content = _content_as_bytes
    storage_obj.set("one", resp)
    cached_resp = storage_obj.get("one")
    assert cached_resp is not None
    if cached_resp is not None:
        assert cached_resp.headers["x-num"] == "one"
        assert cached_resp.status_code == 200
        cached_resp.encoding = "utf8"
        assert cached_resp.text == _content_as_unicode


def test_memory_cache() -> None:
    _test_cache_storage(MemoryCache())


def test_file_cache() -> None:
    fc = FileCache("cache")
    fc.clear()
    _test_cache_storage(fc)
    fc.clear()


def test_sqlite_cache() -> None:
    sc = SqliteCache("cache.db")
    sc.clear()
    _test_cache_storage(sc)
    sc.clear()



class FakeResponse:
    def __init__(
        self,
        url: str,
        code: int,
        content: str | bytes,
        encoding: str = "utf-8",
        headers: dict | None = None,
    ):
        self.url = url
        self.status_code = code
        self.content = content
        self.text = str(content)
        self.encoding = encoding
        self.headers = headers or {}



import abc
import types
import functools
import logging
import re
import os
import glob
import hashlib
import sqlite3
import json

from httpx import Client, Response, Request

log = logging.getLogger("httpx")


def _default_keyfunc(
    method: str,
    url: str,
    params: dict | None = None,
) -> str | None:
    """
    Return a cache key from a given set of request parameters.

    Default behavior is to return a complete URL for all GET
    requests, and None otherwise.
    """
    if method.lower() != "get":
        return None

    return Request(url=url, method=method, params=params).url


def _cache_200s(response: Response) -> bool:
    """
    Check if a given Response object should be cached.

    Default behavior is to only cache responses with a 200 status code.
    """
    return response.status_code == 200


def _cached_request(client: Client, *args, **kwargs):
    method, url = args
    request_key = client._cache_keyfunc(method, url, kwargs["params"])

    # check cache for response
    cached_resp = None
    if request_key and not client._write_only:
        cached_resp = client._cache_storage.get(request_key)

    if cached_resp:
        # resp = cast(CacheResponse, resp_maybe)
        log.info("using cached response request_key=%s", request_key)
        cached_resp.fromcache = True
        resp = cached_resp
    else:
        resp = client._no_cache_request(*args, **kwargs)
        # save to cache if request and response meet criteria
        log.debug("XX %s %s", request_key, client._should_cache(resp))
        if request_key and client._should_cache(resp):
            client._cache_storage.set(request_key, resp)
            log.info("caching response request_key=%s", request_key)
        resp.fromcache = False

    return resp


def make_dev_caching_client(
    *,
    client: Client | None = None,
    cache_storage=None,
    cache_keyfunc=_default_keyfunc,
    should_cache=_cache_200s,
    write_only=False,
):
    """
    Returns an enhanced `httpx.Client` where requests are saved to a
    specified cache.

    This is denoted as a "dev_cache" because it is not intended to be a true
    HTTP cache, respecting cache headers/etc. If you are looking for that
    behavior, there are httpx libraries for that explicit purpose.

    Instead, the purpose of this cache is to make it possible to test scrapers
    locally without making hundreds of redundant requests.

    The strategy is configurable via `cache_keyfunc` and `should_cache`.

    The default strategy is simple:
    cache all GET requests that result in 200s, with no expiry.

    This works well for the case where you have hundreds of pages to scrape
    and want to make scraper adjustments without repeatedly making hits.

    It should *NOT* be used in production without adjusting these rules.
    """
    if client is None:
        client = Client()

    client._cache_storage = cache_storage
    client._cache_keyfunc = cache_keyfunc
    client._should_cache = should_cache
    client._write_only = write_only

    client._no_cache_request = client.request
    client.request = types.MethodType(
        functools.wraps(client.request)(_cached_request), client
    )
    return client


class CacheStorageBase(abc.ABC):
    @abc.abstractmethod
    def get(self, key: str) -> None | Response:
        raise NotImplementedError()

    @abc.abstractmethod
    def set(self, key: str, response: Response) -> None:
        raise NotImplementedError()


class MemoryCache(CacheStorageBase):
    """
    In memory cache for request responses.

    Example:

        make_careful_client(
            cache_storage=MemoryCache(),
        )

    """

    def __init__(self) -> None:
        self.cache: dict[str, Response] = {}

    def get(self, key: str) -> None | Response:
        """Get cache entry for key, or return None."""
        return self.cache.get(key, None)

    def set(self, key: str, response: Response) -> None:
        """Set cache entry for key with contents of response."""
        self.cache[key] = response


class FileCache(CacheStorageBase):
    """
    File-based cache for request responses.

    Parameters:
        cache_dir: directory for storing responses

    Example:

        make_careful_client(
            cache_storage=FileCache("_httpcache/"),
        )

    """

    # TODO: restore?
    # check_last_modified:  set to True to compare last-modified
    #                       timestamp in cached response with value from HEAD request

    # file name escaping inspired by httplib2
    _prefix = re.compile(r"^\w+://")
    _illegal = re.compile(r"[?/:|]+")
    _header_re = re.compile(r"([-\w]+): (.*)")
    _maxlen = 200

    def _clean_key(self, key: str) -> str:
        # strip scheme
        md5 = hashlib.md5(key.encode("utf8")).hexdigest()
        key = self._prefix.sub("", key)
        key = self._illegal.sub(",", key)
        return ",".join((key[: self._maxlen], md5))

    def __init__(self, cache_dir: str, check_last_modified: bool = False):
        # normalize path
        self.cache_dir = os.path.join(os.getcwd(), cache_dir)
        self.check_last_modified = check_last_modified
        # create directory
        if not os.path.isdir(self.cache_dir):
            os.makedirs(self.cache_dir)

    def get(self, orig_key: str) -> None | Response:
        """Get cache entry for key, or return None."""
        key = self._clean_key(orig_key)
        path = os.path.join(self.cache_dir, key)
        resp_headers = {}

        try:
            with open(path, "rb") as f:
                # read lines one at a time
                while True:
                    line = f.readline().decode("utf8").strip("\r\n")
                    # set headers

                    # if self.check_last_modified and re.search(
                    #     "last-modified", line, flags=re.I
                    # ):
                    #     # line contains last modified header
                    #     head_resp = requests.head(orig_key)

                    #     try:
                    #         new_lm = head_resp.headers["last-modified"]
                    #         old_lm = line[line.find(":") + 1 :].strip()
                    #         if old_lm != new_lm:
                    #             # last modified timestamps don't match, need to download again
                    #             return None
                    #     except KeyError:
                    #         # no last modified header present, so redownload
                    #         return None

                    header = self._header_re.match(line)
                    if header:
                        resp_headers[header.group(1)] = header.group(2)
                    else:
                        break
                # everything left is the real content
                resp_content = f.read()

            # status & encoding will be in headers, but are faked
            # need to split spaces out of status to get code (e.g. '200 OK')
            resp = Response(
                status_code=int(resp_headers.pop("status").split(" ")[0]),
                content=resp_content,
                default_encoding=resp_headers.pop("encoding"),
                headers=resp_headers,
            )
            return resp
        except IOError:
            return None

    def set(self, key: str, response: Response) -> None:
        """Set cache entry for key with contents of response."""
        key = self._clean_key(key)
        path = os.path.join(self.cache_dir, key)

        with open(path, "wb") as f:
            status_str = "status: {0}\n".format(response.status_code)
            f.write(status_str.encode("utf8"))
            encoding_str = "encoding: {0}\n".format(response.encoding)
            f.write(encoding_str.encode("utf8"))
            for h, v in response.headers.items():
                # header: value\n
                f.write(h.encode("utf8"))
                f.write(b": ")
                f.write(v.encode("utf8"))
                f.write(b"\n")
            # one blank line
            f.write(b"\n")
            f.write(response.content)

    def clear(self) -> None:
        # only delete things that end w/ a md5, less dangerous this way
        cache_glob = "*," + ("[0-9a-f]" * 32)
        for fname in glob.glob(os.path.join(self.cache_dir, cache_glob)):
            os.remove(fname)


class SqliteCache(CacheStorageBase):
    """
    sqlite cache for request responses.

    Parameters:
        cache_path: path for SQLite database file

    Example:

        make_careful_client(
            cache_storage=SQLiteCache("_cache.db"),
        )
    """

    _columns = ["key", "status", "modified", "encoding", "data", "headers"]

    def __init__(self, cache_path: str, check_last_modified: bool = False):
        self.cache_path = cache_path
        self.check_last_modified = check_last_modified
        self._conn = sqlite3.connect(cache_path)
        self._conn.text_factory = str
        self._build_table()

    def _build_table(self) -> None:
        """Create table for storing request information and response."""
        self._conn.execute(
            """CREATE TABLE IF NOT EXISTS cache
                (key text UNIQUE, status integer, modified text,
                 encoding text, data blob, headers blob)"""
        )

    def set(self, key: str, response: Response) -> None:
        """Set cache entry for key with contents of response."""
        mod = response.headers.pop("last-modified", None)
        status = int(response.status_code)
        rec = (
            key,
            status,
            mod,
            response.encoding,
            response.content,
            json.dumps(dict(response.headers)),
        )
        with self._conn:
            self._conn.execute("DELETE FROM cache WHERE key=?", (key,))
            self._conn.execute("INSERT INTO cache VALUES (?,?,?,?,?,?)", rec)

    def get(self, key: str) -> None | Response:
        """Get cache entry for key, or return None."""
        query = self._conn.execute("SELECT * FROM cache WHERE key=?", (key,))
        rec = query.fetchone()
        if rec is None:
            return None
        rec = dict(zip(self._columns, rec))

        # TODO evaluate/remove?
        # if self.check_last_modified:
        #     if rec["modified"] is None:
        #         return None  # no last modified header present, so redownload

        #     head_resp = requests.head(key)
        #     new_lm = head_resp.headers.get("last-modified", None)
        #     if rec["modified"] != new_lm:
        #         return None

        resp = Response(
            rec["status"],
            content=rec["data"],
            default_encoding=rec["encoding"],
            headers=json.loads(rec["headers"]),
        )
        return resp

    def clear(self) -> None:
        """Remove all records from cache."""
        with self._conn:
            self._conn.execute("DELETE FROM cache")

    def __del__(self) -> None:
        self._conn.close()



from careful.httpx import make_retry_client
from careful.httpx.retries import retry_all_400s_500s
from unittest import mock
from fakeresponse import FakeResponse


def test_retry() -> None:
    client = make_retry_client(attempts=3, wait_seconds=0.001)

    # On the first call return a 500, then a 200
    mock_request = mock.Mock(
        side_effect=[
            FakeResponse("http://dummy/", 500, "failure!"),
            FakeResponse("http://dummy/", 200, "success!"),
        ]
    )

    with mock.patch.object(client, "send", mock_request):
        resp = client.get("http://dummy/")
    assert mock_request.call_count == 2

    # 500 always
    mock_request = mock.Mock(
        return_value=FakeResponse("http://dummy/", 500, "failure!")
    )

    with mock.patch.object(client, "send", mock_request):
        resp = client.get("http://dummy/")
    assert resp.status_code == 500
    assert mock_request.call_count == 4  # try four times


def test_retry_404() -> None:
    client = make_retry_client(
        attempts=3, wait_seconds=0.001, should_retry=retry_all_400s_500s
    )

    # On the first call return a 404, then a 200
    mock_request = mock.Mock(
        side_effect=[
            FakeResponse("http://dummy/", 404, "failure!"),
            FakeResponse("http://dummy/", 200, "success!"),
        ]
    )

    with mock.patch.object(client, "send", mock_request):
        resp = client.get("http://dummy/")  # type: ignore
    assert mock_request.call_count == 2
    assert resp.status_code == 200

    # 404 always
    mock_request = mock.Mock(
        return_value=FakeResponse("http://dummy/", 404, "failure!")
    )

    # four tries
    with mock.patch.object(client, "send", mock_request):
        resp = client.get("http://dummy/")
    assert resp.status_code == 404
    assert mock_request.call_count == 4
    assert resp.status_code == 404


def test_no_retry_404() -> None:
    client = make_retry_client(attempts=3, wait_seconds=0.001)

    # On the first call return a 404, then a 200
    mock_request = mock.Mock(
        side_effect=[
            FakeResponse("http://dummy/", 404, "failure!"),
            FakeResponse("http://dummy/", 200, "success!"),
        ]
    )

    with mock.patch.object(client, "send", mock_request):
        resp = client.get("http://dummy/")  # type: ignore
    assert mock_request.call_count == 1
    assert resp.status_code == 404


# def test_retry_ssl() -> None:
#     s = make_retry_client(retry_attempts=5, retry_wait_seconds=0.001, raise_errors=False)

#     # ensure SSLError is considered fatal even w/ retries
#     with mock.patch.object(requests.Session, "request", mock_sslerror):
#         with pytest.raises(requests.exceptions.SSLError):
#             s.get("http://dummy/", retry_on_404=True)  # type: ignore
#     assert mock_sslerror.call_count == 1



from careful.httpx import make_careful_client, MemoryCache
from unittest import mock
from fakeresponse import FakeResponse


def test_full_careful_client():
    client = make_careful_client(
        retry_attempts=3,
        retry_wait_seconds=0.00001,
        cache_storage=MemoryCache(),
        requests_per_minute=60,
    )

    # On the first call return a 500, then a 200, then a 500 again
    mock_send = mock.Mock(
        side_effect=[
            FakeResponse("http://dummy/", 500, "failure!"),
            FakeResponse("http://dummy/", 200, "success!"),
            FakeResponse("http://dummy/2", 404, "success!"),
        ]
    )

    mock_sleep = mock.Mock()

    # check that sleep is called
    with mock.patch("time.sleep", mock_sleep):
        with mock.patch.object(client, "send", mock_send):
            resp = client.get("http://dummy/")

            # demonstrates a retry
            assert mock_send.call_count == 2
            assert resp.status_code == 200
            # sleep called by retry, not by throttle yet
            assert mock_sleep.call_count == 1

            # demonstrates a cache (no new call)
            resp = client.get("http://dummy/")
            assert mock_send.call_count == 2
            assert resp.status_code == 200
            assert mock_sleep.call_count == 1

            # a new, throttled call (no retry)
            resp = client.get("http://dummy/2")
            assert mock_send.call_count == 3
            assert resp.status_code == 404
            # call was throttled
            assert mock_sleep.call_count == 2



import time
import types
import functools
import logging
from httpx import Client

log = logging.getLogger("httpx")


def _throttle_request(client: Client, *args, **kwargs):
    now = time.time()
    diff = client._request_frequency - (now - client._last_request)
    if diff > 0:
        log.debug("throttled, sleeping for %fs", diff)
        time.sleep(diff)
        client._last_request = time.time()
    else:
        client._last_request = now
    return client._no_throttle_request(*args, **kwargs)


def make_throttled_client(
    *,
    client: Client | None = None,
    requests_per_minute: float = 0,
):
    if requests_per_minute <= 0:
        raise ValueError("requests per minute must be a positive number")

    if client is None:
        client = Client()

    client._last_request = 0.0
    client._requests_per_minute = requests_per_minute
    client._request_frequency = 60.0 / requests_per_minute

    client._no_throttle_request = client.request
    client.request = types.MethodType(
        functools.wraps(client.request)(_throttle_request), client
    )
    return client



import time
import types
import functools
import logging
from httpx import Client, Response, HTTPError

log = logging.getLogger("httpx")


def retry_default_rule(response: Response) -> bool:
    # default behavior is to retry 400s and 500s but not 404s
    return response.status_code >= 400 and response.status_code != 404


def retry_only_500s(response: Response) -> bool:
    return response.status_code >= 500


def retry_all_400s_500s(response: Response) -> bool:
    return response.status_code >= 400


def _retry_request(client: Client, *args, **kwargs):
    # the retry loop
    tries = 0
    exception_raised = None

    while tries <= client._retry_attempts:
        exception_raised = None

        try:
            tries += 1
            resp = client._no_retry_request(*args, **kwargs)

            # break from loop on an accepted response
            if not client._should_retry(resp):
                break

        except HTTPError as e:
            exception_raised = e

            if exception_response := getattr(e, "response", None):
                if not client._should_retry(exception_response):
                    break

        # if we're going to retry, sleep first
        if tries <= client._retry_attempts:
            # twice as long each time
            wait = client._retry_wait_seconds * (2 ** (tries - 1))
            if exception_raised:
                log.info(
                    "exception %s, sleeping for %s seconds before retry #%s",
                    exception_raised,
                    wait,
                    tries,
                )
            else:
                log.info(
                    "response %s, sleeping for %s seconds before retry #%s",
                    resp,
                    wait,
                    tries,
                )
            time.sleep(wait)

    # out of the loop, either an exception was raised or we had a success
    if exception_raised:
        raise exception_raised
    return resp


def make_retry_client(
    *,
    client: Client | None = None,
    attempts: int = 1,
    wait_seconds: float = 10,
    should_retry=retry_default_rule,
):
    if client is None:
        client = Client()
    client._retry_attempts = max(0, attempts)
    client._retry_wait_seconds = wait_seconds
    client._should_retry = should_retry

    client._no_retry_request = client.request
    client.request = types.MethodType(
        functools.wraps(client.request)(_retry_request), client
    )

    return client
